@@ -30,7 +30,7 @@ use Config;
 use Digest::SHA;
 use Data::Dumper;
 use IPC::Open2;
-use File::Path qw(:DEFAULT make_path);
+use File::Path qw(make_path);
 use File::Basename;
 use Dpkg::Control::Hash;
 use Debian::Dgit::ExitStatus;
@@ -53,7 +53,7 @@ BEGIN {
                       stat_exists link_ltarget rename_link_xf
 		      hashfile
                       fail failmsg ensuredir ensurepath
-		      must_getcwd executable_on_path
+		      must_getcwd rmdir_r executable_on_path
                       waitstatusmsg failedcmd_waitstatus
 		      failedcmd_report_cmd failedcmd
                       runcmd runcmd_quieten
@@ -112,9 +112,9 @@ our $branchprefix = 'dgit';
 our $series_filename_re = qr{(?:^|\.)series(?!\n)$}s;
 our $extra_orig_namepart_re = qr{[-0-9a-zA-Z]+};
 our $orig_f_comp_re = qr{orig(?:-$extra_orig_namepart_re)?};
-our $orig_f_sig_re = '\\.(?:asc|gpg|pgp)';
-our $tarball_f_ext_re = "\\.tar(?:\\.\\w+)?(?:$orig_f_sig_re)?";
-our $orig_f_tail_re = "$orig_f_comp_re$tarball_f_ext_re";
+our $orig_f_sig_re = qr{\.(?:asc|gpg|pgp)};
+our $tarball_f_ext_re = qr{\.tar(?:\.\w+)?(?:$orig_f_sig_re)?};
+our $orig_f_tail_re = qr{$orig_f_comp_re$tarball_f_ext_re};
 our $git_null_obj = '0' x 40;
 our $ffq_refprefix = 'ffq-prev';
 our $gdrlast_refprefix = 'debrebase-last';
@@ -808,6 +808,49 @@ sub changedir ($) {
     chdir $newdir or confess "chdir: $newdir: $!";
 }
 
+sub rmdir_r ($) {
+    my ($dir) = @_;
+    # Removes the whole subtree $dir (which need not exist), or dies.
+    #
+    # We used to use File::Path::remove_tree (via ::rmtree) but its
+    # error handling and chmod behaviour is very complex and confusing.
+    # For example:
+    #   - It chdirs, and then tries to chdir back.
+    #     With our $SIG{__WARN__} setting to die, this can, in combination
+    #     with eval, cause execution to continue in an unexpected cwd!
+    #   - Without the `safe` option it chmods things.  We never want that.
+    #   - *With* the `safe` option it appears to silently skip things
+    #     it would want to chdir.
+    #   - The error handling with the errors option is rather janky.
+
+    # Don't fork/exec rm if we don't have to.
+    return unless stat_exists $dir;
+
+    # We don't use runcmd because we want to capture errors from rmdir
+    # in $@ so that if we eval a rmdir_r, the right things happen.
+    # even cmdoutput_errok is too cooked.
+    my @cmd = (qw(rm -rf --), $dir);
+    debugcmd '+', @cmd;
+    my $child = open P, "-|" // confess $!;
+    if (!$child) {
+	open STDERR, ">& STDOUT" or die $!;
+	exec @cmd or die "exec $cmd[0]: $!";
+    }
+    $!=0; $?=0;
+    my $errs;
+    { local $/ = undef; $errs = <P>; }
+    confess "$!" if P->error;
+
+    return if close P && !stat_exists $dir;
+
+    chomp $errs;
+    $errs =~ s{\n}{; }g;
+    $errs ||= 'no error messages';
+
+    die f_ "failed to remove directory tree %s: rm -rf: %s; %s\n",
+      $dir, waitstatusmsg, $errs;
+}
+
 sub git_slurp_config_src ($) {
     my ($src) = @_;
     # returns $r such that $r->{KEY}[] = VALUE
@@ -1118,7 +1161,7 @@ END
 #    playtree_setup
 #            # ^ call in some (perhaps trivial) subdir of playground
 #
-#    rmtree playground
+#    rmdir_r playground
 
 # ----- maindir -----
 
@@ -1166,7 +1209,7 @@ sub ensure_a_playground ($) {
 sub fresh_playground ($) {
     my ($spc) = @_;
     $spc = ensure_a_playground_parent $spc;
-    rmtree $spc;
+    rmdir_r $spc;
     mkdir $spc or fail f_
 	"failed to mkdir the playground %s: %s", $spc, $!;
     return $spc;
